"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProtoClient = void 0;
const events_1 = require("events");
const Protobuf = __importStar(require("protobufjs"));
const grpc_js_1 = require("@grpc/grpc-js");
const ProtoRequest_1 = require("./ProtoRequest");
/**
 * Shorthand wrapper utilities for gRPC client calls
 */
class ProtoClient extends events_1.EventEmitter {
    /**
     * Client settings
     */
    clientSettings = {
        endpoint: `0.0.0.0:8080`,
    };
    /**
     * Protobuf parsing settings
     */
    protoSettings = {};
    /**
     * Internal proto conversion options, for referencing
     */
    protoConversionOptions = {
        longs: Number,
        enums: String,
        defaults: false,
        oneofs: true,
    };
    /**
     * List of currently active requests
     */
    activeRequests = [];
    /**
     * Middleware to run before each request
     */
    middleware = [];
    /**
     * Root protobuf object for referencing
     */
    root;
    /**
     * Client endpoints list
     */
    clients = [];
    /**
     * Shorthand wrapper utilities for gRPC client calls
     * @param options Configuration options for the client
     */
    constructor(options) {
        super();
        if (options?.clientSettings) {
            this.configureClient(options.clientSettings);
        }
        if (options?.protoSettings) {
            this.configureProtos(options.protoSettings);
        }
        if (options?.conversionOptions) {
            this.configureConversionOptions(options.conversionOptions);
        }
    }
    /**
     * Configures settings for client connection
     * @param clientSettings Settings for client connection
     */
    configureClient(clientSettings) {
        this.clientSettings = clientSettings;
        this.close();
        const endpoints = Array.isArray(clientSettings.endpoint)
            ? clientSettings.endpoint
            : [clientSettings.endpoint];
        this.clients = endpoints.map((endpoint) => {
            if (typeof endpoint === "string") {
                return {
                    match: this.generateEndpointMatcher(),
                    endpoint: { address: endpoint },
                    client: new grpc_js_1.Client(endpoint, grpc_js_1.credentials.createInsecure()),
                };
            }
            // Build out credentials (secure/insecure) for this endpoint
            let creds;
            if (endpoint.credentials &&
                endpoint.credentials instanceof grpc_js_1.ChannelCredentials) {
                creds = endpoint.credentials;
            }
            else if (endpoint.credentials) {
                creds = grpc_js_1.credentials.createSsl(endpoint.credentials.rootCerts, endpoint.credentials.privateKey, endpoint.credentials.certChain, endpoint.credentials.verifyOptions);
            }
            else {
                creds = grpc_js_1.credentials.createInsecure();
            }
            // Combined matcher with gRPC client connection
            return {
                endpoint,
                match: this.generateEndpointMatcher(endpoint.match),
                client: new grpc_js_1.Client(endpoint.address, creds, endpoint.clientOptions),
            };
        });
    }
    /**
     * Configures options for reading/writing proto definitions
     * @param protoSettings Proto reading/writing settings
     */
    configureProtos(protoSettings) {
        this.protoSettings = protoSettings;
        if (this.protoSettings.files) {
            this.root = new Protobuf.Root();
            this.root
                .loadSync(this.protoSettings.files, this.protoSettings.parseOptions)
                .resolveAll();
        }
        else if (this.protoSettings.root) {
            this.root = this.protoSettings.root;
        }
        else {
            throw new Error(`Must define either a root protobuf object, or path to proto files`);
        }
    }
    /**
     * Configures conversion options when decoding proto responses
     * @param options Conversion options
     */
    configureConversionOptions(options) {
        this.protoConversionOptions = options;
    }
    /**
     * Fetches the endpoint matching client instance
     * throwing an error if not yet configured
     * @param request Active request looking for it's client endpoint
     */
    getClient(request) {
        if (!this.clients.length) {
            throw new Error(`ProtoClient is not yet configured`);
        }
        // Find the client which matches the request
        for (const client of this.clients) {
            if (client.match(request.method, request)) {
                return client.client;
            }
        }
        // Raise exception when there is no matching client
        throw new Error(`Service method '${request.method}' has no configured endpoint`);
    }
    /**
     * Fetches the current proto root object, throwing an error if not yet configured
     */
    getRoot() {
        if (!this.root) {
            throw new Error(`ProtoClient protos are not yet configured`);
        }
        return this.root;
    }
    /**
     * Adds function to the middleware stack
     * @param middleware Middleware to run before each request
     */
    useMiddleware(middleware) {
        this.middleware.push(middleware);
    }
    // Actual makeUnaryRequest for the overloads above
    async makeUnaryRequest(method, data, requestOptions) {
        const request = await this.generateRequest(method, "makeUnaryRequest" /* RequestMethodType.UnaryRequest */, requestOptions);
        return request.makeUnaryRequest(data);
    }
    // Actual makeClientStreamRequest for the overloads above
    async makeClientStreamRequest(method, writerSandbox, requestOptions) {
        const request = await this.generateRequest(method, "makeClientStreamRequest" /* RequestMethodType.ClientStreamRequest */, requestOptions);
        return request.makeClientStreamRequest(writerSandbox);
    }
    // Actual makeServerStreamRequest for the overloads above
    async makeServerStreamRequest(method, data, streamReader, requestOptions) {
        const request = await this.generateRequest(method, "makeServerStreamRequest" /* RequestMethodType.ServerStreamRequest */, requestOptions);
        return request.makeServerStreamRequest(data, streamReader);
    }
    // Actual makeBidiStreamRequest for the overloads above
    async makeBidiStreamRequest(method, writerSandbox, streamReader, requestOptions) {
        const request = await this.generateRequest(method, "makeBidiStreamRequest" /* RequestMethodType.BidiStreamRequest */, requestOptions);
        return request.makeBidiStreamRequest(writerSandbox, streamReader);
    }
    // Actual abortRequests for the overloads above
    abortRequests(match) {
        const matcher = this.generateEndpointMatcher(match);
        this.activeRequests.forEach((request) => {
            if (matcher(request.method, request)) {
                request.abort();
            }
        });
    }
    /**
     * Aborts all active requests and closes all gRPC connections
     */
    close() {
        this.abortRequests(() => true);
        this.clients.forEach((client) => client.client.close());
        this.clients = [];
    }
    /**
     * Manages request within the active requests list
     * @param method Method name for the request
     * @param requestMethodType Method type
     */
    async generateRequest(method, requestMethodType, requestOptions) {
        const request = new ProtoRequest_1.ProtoRequest(this, method, requestMethodType, requestOptions);
        request.on("aborted", () => this.removeActiveRequest(request));
        request.on("end", () => this.removeActiveRequest(request));
        this.activeRequests.push(request);
        // Run middleware
        try {
            for (const middleware of this.middleware) {
                await middleware(request, this);
            }
        }
        catch (e) {
            if (e instanceof Error) {
                request.error = e;
            }
            else if (typeof e === "string") {
                request.error = new Error(e);
            }
            else {
                request.error = new Error(`Unknown Middleware Error: ${e}`);
            }
            this.removeActiveRequest(request);
            request.emit("end", request);
            throw request.error;
        }
        this.emit("request", request);
        return request;
    }
    /**
     * Removes request from the active request list
     * @param request Request to be removed
     */
    removeActiveRequest(request) {
        const index = this.activeRequests.indexOf(request);
        if (index > -1) {
            this.activeRequests.splice(index, 1);
        }
    }
    /**
     * Generates a function for matching namespaces to requests
     * @param match Multi-type endpoint matcher to convert
     * @returns Matcher function
     */
    generateEndpointMatcher(match) {
        if (!match) {
            return () => true;
        }
        else if (typeof match === "string") {
            if (/\.\*$/.test(match)) {
                const matchNamespace = match.replace(/\*$/, "");
                return (method) => method.startsWith(matchNamespace);
            }
            else {
                return (method) => method === match;
            }
        }
        else if (match instanceof RegExp) {
            return (method) => match.test(method);
        }
        else {
            return match;
        }
    }
}
exports.ProtoClient = ProtoClient;
//# sourceMappingURL=ProtoClient.js.map